#!/usr/bin/env python3

import sys
import os
import subprocess
import ipaddress
import json

PLAYBOOKS_DIR="/usr/share/xivo/ansible/playbooks"
INSTALL_PLAYBOOK=PLAYBOOKS_DIR + "/xivo-install.yml"


class XivoModule:
    def __init__(self, name, label, descr):
        self.name = name
        self.label = label
        self.descr = descr

class ServerType:
    def __init__(self, modules):
        self.modules = modules

    def descr(self):
        return "+".join([ m.descr for m in self.modules ])

    def get_dict(self):
        return {
            'modules': [ m.name for m in self.modules ],
            'labels': [ m.label for m in self.modules ]
        }
    

class DeploymentType:
    def __init__(self, servers):
        for s in servers:
            if len(s) != 2:
                raise TypeError()
            
            if not isinstance(s[0], ServerType):
                raise TypeError()
        
        self.servers = servers
        
        
    def descr(self):
        cnt = self.srv_count()
        srv = "{} server{}, ".format(cnt, "" if cnt == 1 else "s")
        return srv + " / ".join([ s[0].descr() for s in self.servers ])

    def srv_count(self):
        return len(self.servers)

    def get_dict(self):
        servers = [ s[0].get_dict() for s in self.servers ]
        for i in range(self.srv_count()):
            servers[i]['manager'] = self.servers[i][1]
        
        return servers
        
        

class Deployment:
    def __init__(self, depl_type):
        
        self.depl_type = depl_type
        self.ips = [None] * self.srv_count()
    
    def srv_count(self):
        return self.depl_type.srv_count()
    
    def set_ip(self, srv_index, ip):
        if srv_index < 0 or srv_index >= self.srv_count():
            raise Error
        
        self.ips[srv_index] = ip
        
        
    def get_dict(self, exclude_main = False):
        servers = self.depl_type.get_dict()
        for i in range(self.srv_count()):
            servers[i]['ip'] = self.ips[i]
        if exclude_main:
            servers.pop(0)    
        return { 'servers': servers }

        
class Whiptail:
    
    path = "whiptail"
    
    def __init__(self, title):
        self.title = title
    
    def set_path(self, path):
        self.path = path
    
    def run(self, args, cancelcheck = None):
        cmd = [self.path] + ["--title", self.title] + args
        res = subprocess.run(cmd, stderr=subprocess.PIPE)
        
        # Cancel or No button pressed - ret == 1
        # Esc key pressed - ret == 255 and no output on stderr
        # whiptaill error - ret == 255 and usage printed to stderr 

        if res.returncode == 1 or res.returncode == 255 and res.stderr == '':
            if cancelcheck:
                cancelcheck()
            return None
        elif res.returncode != 0:
            raise RuntimeError("Whiptail error")
        elif res.stderr == b'':
            return ['']
        else:
            return res.stderr.decode('ascii').split("\n")
        

class WhiptailRadiolist(Whiptail):
    
    def __init__(self, title, items, text = "", width = 60):
        super().__init__(title)
        self.text = text
        self.width = width
        self.items = items
        super().__init__(title)
        
        
    def run(self, args=[], cancelcheck = None):
        listh = len(self.items)
        h = listh + 5
        
        fullargs = args + ["--radiolist", self.text, str(h), str(self.width), str(listh) ]
        for i in self.items:
            fullargs += [ i[0], i[1], "off" ]
        
        
        ret =  super().run(fullargs, cancelcheck)
        return None if ret is None else ret[0]


class WhiptailInputbox(Whiptail):
    
    def __init__(self, title, text = "", height = 10, width = 60, init = ""):
        super().__init__(title)
        self.text = text
        self.height = height
        self.width = width
        self.init = init
        super().__init__(title)
        
        
    def run(self, args=[], cancelcheck = None):
        
        fullargs = args + [
            "--inputbox", 
            self.text, 
            str(self.height),
            str(self.width),
            self.init
        ]
        
        ret = super().run(fullargs, cancelcheck)
        return None if ret is None else ret[0]

class WhiptailMsgbox(Whiptail):
    
    def __init__(self, title, msg, height = 7, width = 60):
        self.msg = msg
        self.height = height
        self.width = width
        
        super().__init__(title)

    def run(self, args=[]):
        fullargs = args + ["--msgbox", self.msg, str(self.height), str(self.width)]
        super().run(fullargs)


class WhiptailYesno(Whiptail):
    
    def __init__(self, title, msg, height = 7, width = 60):
        self.msg = msg
        self.height = height
        self.width = width
        
        super().__init__(title)

    def run(self, args=[]):
        fullargs = args + ["--yesno", self.msg, str(self.height), str(self.width)]
        res = super().run(fullargs)
        
        return False if res is None else True
        

module_xivo = XivoModule("xivo", "xivo_main", "Xivo")
module_xuc = XivoModule("xuc", "xivo_xuc", "XuC")        
module_rep = XivoModule("rep", "xivo_rep", "Reporting")
module_rec = XivoModule("rec", "xivo_rec", "Recording")

server_type_xivo = ServerType([module_xivo])
server_type_xivo_xuc = ServerType([module_xivo, module_xuc, module_rep, module_rec])

server_type_xcc_all = ServerType([module_xivo, module_xuc, module_rep, module_rec])
server_type_xcc_xuc = ServerType([module_xuc])
server_type_xcc_xuc_rep = ServerType([module_xuc, module_rep])
server_type_xcc_rep = ServerType([module_rep])
server_type_xcc_rec = ServerType([module_rec])

depl_types = {
    "xivo_only": DeploymentType([(server_type_xivo, True)]),
#    "xivouc1": DeploymentType([(server_type_xivo_xuc, True)]),
#    "xivouc2": DeploymentType([(server_type_xivo, True), (server_type_xuc, False)]),
    "xivocc2": DeploymentType([(server_type_xivo, True),(server_type_xcc_all, False)]),
    "xivocc3": DeploymentType([(server_type_xivo, True),(server_type_xcc_xuc_rep, True), (server_type_xcc_rec, True)]),
    "xivocc4": DeploymentType([(server_type_xivo, True),(server_type_xcc_xuc, True), (server_type_xcc_rep, True), (server_type_xcc_rec, False)]) 
}


def proc_cancel():
    w = WhiptailYesno("Cancel", "Do you want to cancel installation?")
    res = w.run()
    
    if res:
        exit(0)
    

def show_msg(title, msg):
    w = WhiptailMsgbox(title, msg)
    w.run()


def show_error(msg):
    show_msg("Error", msg)


def select_deployment_type():
    items = [ (k, v.descr()) for k, v in depl_types.items() ]
    w = WhiptailRadiolist("Select deployment type", items, width=90)
    
    res = ""
    while True:
        res = w.run(["--notags"], proc_cancel)
        if res is None:
            continue
        
        if res == '':
            show_msg("Error", "Nothing selected")
            continue
            
        return res

def get_server_ip(server_type, used_ips, ip = ""):
    title = "Server IP"
    txt = "IP address of server with {}".format(server_type.descr())

    input = ''

    while True:
        w = WhiptailInputbox(title, txt, 10, 60, input)
        input = w.run(cancelcheck = proc_cancel)

        if input is None:
            input = ''
            continue

        try: 
            ip = ipaddress.IPv4Address(input)
        except ipaddress.AddressValueError: 
            show_error("Invalid IP")
            continue

        ipstr = str(ip)
        if ipstr in used_ips:
            show_error("IP already used")
            continue

        return ipstr         

def get_server_ips(depl):
    used = []
    i = 0
    for s in depl.depl_type.servers:
        ip = get_server_ip(s[0], used)
        if ip is None:
            return None
        used.append(ip)
        depl.set_ip(i, ip)
        i = i + 1

def get_inventory_str(depl, exclude_main=False):
    d = depl.get_dict(exclude_main)
    ips = [ s['ip'] for s in d['servers'] ]
    return ','.join(ips)

def install_all(depl):

    servers_str = json.dumps(depl.get_dict())
    inventory = get_inventory_str(depl)

    cmd = ['ansible-playbook', '-i', inventory, '-e', servers_str, INSTALL_PLAYBOOK, '-e', 'ansible_python_interpreter=python3' ]
    subprocess.run(cmd)

t = select_deployment_type()
depl = Deployment(depl_types[t])
get_server_ips(depl)
install_all(depl)
